@@ -42,7 +42,8 @@ int main (int argc, char **argv)
   uint64_t freebytes;
   char *storage_description;
   char *volume_label;
-  char *owner;
+  char *friendlyname;
+  char *syncpartner;
   char *sectime;
   char *devcert;
   uint16_t *filetypes;
@@ -80,13 +81,20 @@ int main (int argc, char **argv)
   LIBMTP_Dump_Device_Info(device);
   
   printf("MTP-specific device properties:\n");
-  // The owner name
-  owner = LIBMTP_Get_Ownername(device);
-  if (owner == NULL) {
-    printf("   Owner name: (NULL)\n");
+  // The friendly name
+  friendlyname = LIBMTP_Get_Friendlyname(device);
+  if (friendlyname == NULL) {
+    printf("   Friendly name: (NULL)\n");
   } else {
-    printf("   Owner name: %s\n", owner);
-    free(owner);
+    printf("   Friendly name: %s\n", friendlyname);
+    free(friendlyname);
+  }
+  syncpartner = LIBMTP_Get_Syncpartner(device);
+  if (syncpartner == NULL) {
+    printf("   Synchronization partner: (NULL)\n");
+  } else {
+    printf("   Synchronization partner: %s\n", syncpartner);
+    free(syncpartner);
   }
 
   // Some storage info
@@ -121,7 +129,8 @@ int main (int argc, char **argv)
     printf("   Battery level %d of %d (%d%%)\n",currbattlevel, maxbattlevel, 
 	   (int) ((float) currbattlevel/ (float) maxbattlevel * 100.0));
   } else {
-    printf("   Error getting battery info...\n");
+    // Silently ignore. Some devices does not support getting the 
+    // battery level.
   }
 
   ret = LIBMTP_Get_Supported_Filetypes(device, &filetypes, &filetypes_len);
@@ -148,7 +157,7 @@ int main (int argc, char **argv)
     free(devcert);
   }
 
-  // Try to get device info XML file...
+  // Try to get Media player device info XML file...
   files = LIBMTP_Get_Filelisting(device);
   if (files != NULL) {
     LIBMTP_file_t *file, *tmp;
