@@ -55,55 +55,122 @@
  * this list by stealing from everyone I know.
  */
 static const LIBMTP_device_entry_t mtp_device_table[] = {
-  /* Creative devices */
-  { "Creative Zen Vision", 0x041e, 0x411f },
-  { "Creative Portable Media Center", 0x041e, 0x4123 },
-  { "Creative Zen Xtra (MTP mode)", 0x041e, 0x4128 },
-  { "Second generation Dell DJ", 0x041e, 0x412f },
-  { "Creative Zen Micro (MTP mode)", 0x041e, 0x4130 },
-  { "Creative Zen Touch (MTP mode)", 0x041e, 0x4131 },
-  { "Dell Pocket DJ (MTP mode)", 0x041e, 0x4132 },
-  { "Creative Zen Sleek (MTP mode)", 0x041e, 0x4137 },
-  { "Creative Zen MicroPhoto", 0x041e, 0x413c },
-  { "Creative Zen Sleek Photo", 0x041e, 0x413d },
-  { "Creative Zen Vision:M", 0x041e, 0x413e },
-  /* From libgphoto2 */
-  { "Samsung:YH-820", 0x04e8, 0x502e },
-  /* Contributed by polux2001@users.sourceforge.net */
-  { "Samsung YH-925", 0x04e8, 0x502f },
-  /* Contributed by anonymous person on SourceForge */
-  { "Samsung YP-T7J", 0x04e8, 0x5047 },
-  /* From a rouge .INF file */
-  { "Samsung YH-999 Portable Media Center", 0x04e8, 0x5a0f },
-  { "Intel Bandon Portable Media Center", 0x045e, 0x00c9 },
-  { "iRiver Portable Media Center", 0x1006, 0x4002 },
-  { "iRiver Portable Media Center", 0x1006, 0x4003 },
-  /* From Mark Veinot */
-  { "JVC Alneo XA-HD500", 0x04f1, 0x6105 },
-  /* From libgphoto2 source */
-  { "Philipps HDD6320", 0x0471, 0x01eb },
-  { "Philipps HDD6320 2", 0x0471, 0x014b },
-  /* Anonymous SourceForge user */
-  { "Philipps HDD1630/17", 0x0471, 0x014c },
-  /* Reported by tangent_@users.sourceforge.net */
-  { "SanDisk Sansa c150", 0x0781, 0x7410 },
-  /* From libgphoto2 source */
-  { "SanDisk Sansa e200", 0x0781, 0x7420 },
-  /* Reported by gonkflea@users.sourceforge.net */
-  { "SanDisk Sansa e260", 0x0781, 0x7420 },
-  /* From libgphoto2 source */
-  { "iRiver T10", 0x4102, 0x1113 },
-  { "iRiver T20 FM", 0x4102, 0x1114 },
-  { "iRiver U10", 0x4102, 0x1116 },
-  { "iRiver T10", 0x4102, 0x1117 },
-  { "iRiver T20", 0x4102, 0x1118 },
-  { "iRiver T30", 0x4102, 0x1119 },
-  /* Reported by Scott Call */
-  { "iRiver H10", 0x4102, 0x2101 },
-  { "iRiver H10", 0x4102, 0x2102 },
-  { "Dell DJ Itty", 0x413c, 0x4500 },
-  { "Toshiba Gigabeat MEGF-40", 0x0930, 0x0009 },
-  { "Toshiba Gigabeat", 0x0930, 0x000c }
+  
+  /*
+   * Creative Technology
+   */
+  { "Creative Zen Vision", 0x041e, 0x411f, DEVICE_FLAG_NONE },
+  { "Creative Portable Media Center", 0x041e, 0x4123, DEVICE_FLAG_NONE },
+  { "Creative Zen Xtra (MTP mode)", 0x041e, 0x4128, DEVICE_FLAG_NONE },
+  { "Second generation Dell DJ", 0x041e, 0x412f, DEVICE_FLAG_NONE },
+  { "Creative Zen Micro (MTP mode)", 0x041e, 0x4130, DEVICE_FLAG_NONE },
+  { "Creative Zen Touch (MTP mode)", 0x041e, 0x4131, DEVICE_FLAG_NONE },
+  { "Dell Pocket DJ (MTP mode)", 0x041e, 0x4132, DEVICE_FLAG_NONE },
+  { "Creative Zen Sleek (MTP mode)", 0x041e, 0x4137, DEVICE_FLAG_NONE },
+  { "Creative Zen MicroPhoto", 0x041e, 0x413c, DEVICE_FLAG_NONE },
+  { "Creative Zen Sleek Photo", 0x041e, 0x413d, DEVICE_FLAG_NONE },
+  { "Creative Zen Vision:M", 0x041e, 0x413e, DEVICE_FLAG_NONE },
+  // Reported by marazm@o2.pl
+  { "Creative Zen V", 0x041e, 0x4150, DEVICE_FLAG_NONE },
+  // Reported by Darel on the XNJB forums
+  { "Creative Zen V plus", 0x041e, 0x4152, DEVICE_FLAG_NONE },
+
+  /*
+   * Samsung
+   * We suspect that more of these are dual mode.
+   */
+  // From libgphoto2
+  { "Samsung YH-820", 0x04e8, 0x502e, DEVICE_FLAG_NONE },
+  // Contributed by polux2001@users.sourceforge.net
+  { "Samsung YH-925", 0x04e8, 0x502f, DEVICE_FLAG_NONE },
+  // Contributed by anonymous person on SourceForge
+  { "Samsung YP-T7J", 0x04e8, 0x5047, DEVICE_FLAG_NONE },
+  // Reported by cstrickler@gmail.com
+  { "Samsung YP-U2J (YP-U2JXB/XAA)", 0x04e8, 0x5054, DEVICE_FLAG_NONE },
+  // Reported by Andrew Benson
+  { "Samsung YP-F2J", 0x04e8, 0x5057, DEVICE_FLAG_DUALMODE },
+  // From a rouge .INF file
+  { "Samsung YH-999 Portable Media Center", 0x04e8, 0x5a0f, DEVICE_FLAG_NONE },
+
+  /*
+   * Intel
+   */
+  { "Intel Bandon Portable Media Center", 0x045e, 0x00c9, DEVICE_FLAG_NONE },
+
+  /*
+   * JVC
+   */
+  // From Mark Veinot
+  { "JVC Alneo XA-HD500", 0x04f1, 0x6105, DEVICE_FLAG_NONE },
+
+  /*
+   * Philipps
+   */
+  // From libgphoto2 source
+  { "Philipps HDD6320", 0x0471, 0x01eb, DEVICE_FLAG_NONE },
+  { "Philipps HDD6320 2", 0x0471, 0x014b, DEVICE_FLAG_NONE },
+  // Anonymous SourceForge user
+  { "Philipps HDD1630/17", 0x0471, 0x014c, DEVICE_FLAG_NONE },
+
+  /*
+   * SanDisk
+   */
+  // Reported by Brian Robison
+  { "SanDisk Sansa m240", 0x0781, 0x7400, DEVICE_FLAG_NONE },
+  // Reported by tangent_@users.sourceforge.net
+  { "SanDisk Sansa c150", 0x0781, 0x7410, DEVICE_FLAG_NONE },
+  // From libgphoto2 source
+  { "SanDisk Sansa e200", 0x0781, 0x7420, DEVICE_FLAG_NONE },
+  // Reported by gonkflea@users.sourceforge.net
+  { "SanDisk Sansa e260", 0x0781, 0x7420, DEVICE_FLAG_NONE },
+
+  /*
+   * iRiver
+   */
+  { "iRiver Portable Media Center", 0x1006, 0x4002, DEVICE_FLAG_NONE },
+  { "iRiver Portable Media Center", 0x1006, 0x4003, DEVICE_FLAG_NONE },
+  // From libgphoto2 source
+  { "iRiver T10", 0x4102, 0x1113, DEVICE_FLAG_NONE },
+  { "iRiver T20 FM", 0x4102, 0x1114, DEVICE_FLAG_NONE },
+  { "iRiver U10", 0x4102, 0x1116, DEVICE_FLAG_NONE },
+  { "iRiver T10", 0x4102, 0x1117, DEVICE_FLAG_NONE },
+  { "iRiver T20", 0x4102, 0x1118, DEVICE_FLAG_NONE },
+  { "iRiver T30", 0x4102, 0x1119, DEVICE_FLAG_NONE },
+  // Reported by David Wolpoff
+  { "iRiver T10 2GB", 0x4102, 0x1120, DEVICE_FLAG_NONE },
+  // Reported by Adam Torgerson
+  { "iRiver Clix", 0x4102, 0x112a, DEVICE_FLAG_NONE },
+  // Reported by Scott Call
+  { "iRiver H10 20GB", 0x4102, 0x2101, DEVICE_FLAG_NONE },
+  { "iRiver H10", 0x4102, 0x2102, DEVICE_FLAG_NONE },
+
+  /*
+   * Dell
+   */
+  { "Dell DJ Itty", 0x413c, 0x4500, DEVICE_FLAG_NONE },
+  
+  /*
+   * Toshiba
+   */
+  { "Toshiba Gigabeat MEGF-40", 0x0930, 0x0009, DEVICE_FLAG_NONE },
+  { "Toshiba Gigabeat", 0x0930, 0x000c, DEVICE_FLAG_NONE },
+  // Reported by Rob Brown
+  { "Toshiba Gigabeat P10", 0x0930, 0x0011, DEVICE_FLAG_NONE },
+
+  /*
+   * Archos
+   */
+  // Reported by gudul1@users.sourceforge.net
+  { "Archos 104 (MTP mode)", 0x0e79, 0x120a, DEVICE_FLAG_NONE },
+
+  /*
+   * Dunlop (OEM of EGOMAN ltd?) reported by Nanomad
+   * This unit is falsely detected as USB mass storage in Linux
+   * so special care is needed to blacklist the device in kernel
+   * or similar to get it properly userspace:d.
+   */
+  { "Dunlop MP3 player 1GB / EGOMAN MD223AFD", 0x10d6, 0x2200, DEVICE_FLAG_UNLOAD_DRIVER}
+  
 };
 static const int mtp_device_table_size = sizeof(mtp_device_table) / sizeof(LIBMTP_device_entry_t);
 
@@ -174,16 +241,12 @@ static struct usb_device *probe_usb_bus_for_mtp_devices(void)
 	usb_close(devh);
 	continue;
       }
-
-      if (ret > 0) {
-	// printf("Device: VID %04x/PID %04x: responds to special descriptor call...\n",
-	//       dev->descriptor.idVendor,
-	//       dev->descriptor.idProduct);
-	// data_dump_ascii (stdout, buf, ret, 0);
-      }
-
+      
+      // It is atleast 10 bytes...
       if (!((buf[2] == 'M') && (buf[4]=='S') && (buf[6]=='F') && (buf[8]=='T'))) {
-	// printf("This is not a Microsoft MTP descriptor...\n");
+	printf("This is not a Microsoft MTP descriptor...\n");
+	printf("Device response to read device property 0xee:\n");
+	data_dump_ascii (stdout, buf, ret, 0);
 	usb_close(devh);
 	continue;
       }
@@ -194,13 +257,21 @@ static struct usb_device *probe_usb_bus_for_mtp_devices(void)
       if (ret == -1) {
 	//printf("Decice could not respond to control message 1.\n");
 	usb_close(devh);
-	// Return the device anyway.
-	return dev;
+	continue;
       }
       
-      if (ret > 0) {
-	//printf("Device response to control message 1:\n");
-	//data_dump_ascii (stdout, buf, ret, 0);
+      if (ret > 0x15) {
+	if ((buf[0x12] != 'M') || (buf[0x13] != 'T') || (buf[0x14] != 'P')) {
+	  printf("The device has a Microsoft device descriptor, but it's not MTP.\n");
+	  printf("This is not an MTP device. Presumable it is USB mass storage\n");
+	  printf("with some additional Janus (DRM) support.\n");
+	  printf("Device response to control message 1:\n");
+	  data_dump_ascii (stdout, buf, ret, 0);
+	  continue;
+	}
+      } else {
+	// Not MTP or broken
+	continue;
       }
 
       ret = usb_control_msg (devh, USB_ENDPOINT_IN|USB_RECIP_DEVICE|USB_TYPE_VENDOR, 
@@ -208,15 +279,20 @@ static struct usb_device *probe_usb_bus_for_mtp_devices(void)
       if (ret == -1) {
 	//printf("Device could not respond to control message 2.\n");
 	usb_close(devh);
-	// Return the device anyway.
+	// Return the device anyway, it said previously it was MTP, right?
 	return dev;
       }
       
-      if (ret > 0) {
-	//printf("Device response to control message 2:\n");
-	//data_dump_ascii (stdout, buf, ret, 0);
+      if (ret > 0x15) {
+	if ((buf[0x12] != 'M') || (buf[0x13] != 'T') || (buf[0x14] != 'P')) {
+	  printf("This device does not respond with MTP characteristics on\n");
+	  printf("the second device property read (0x05), but will be regarded\n");
+	  printf("as MTP anyway.\n");
+	  printf("Device response to control message 2:\n");
+	  data_dump_ascii (stdout, buf, ret, 0);
+	}
       }
-
+      
       usb_close(devh);
       // We can return the device here, it will be the first. 
       // If it was not MTP, the loop continues before it reaches this point.
@@ -273,15 +349,17 @@ void dump_usbinfo(PTP_USB *ptp_usb)
   printf("   bDeviceClass: %d\n", dev->descriptor.bDeviceClass);
   printf("   bDeviceSubClass: %d\n", dev->descriptor.bDeviceSubClass);
   printf("   bDeviceProtocol: %d\n", dev->descriptor.bDeviceProtocol);
-  printf("   bMaxPacketSize0: %d\n", dev->descriptor.bMaxPacketSize0);
   printf("   idVendor: %04x\n", dev->descriptor.idVendor);
   printf("   idProduct: %04x\n", dev->descriptor.idProduct);
+  printf("   IN endpoint maxpacket: %d bytes\n", ptp_usb->inep_maxpacket);
+  printf("   OUT endpoint maxpacket: %d bytes\n", ptp_usb->outep_maxpacket);
+  printf("   Device flags: 0x%08x\n", ptp_usb->device_flags);
   // TODO: add in string dumps for iManufacturer, iProduct, iSerialnumber...
 }
 
 
 // Based on same function on library.c in libgphoto2
-#define CONTEXT_BLOCK_SIZE	100000
+#define CONTEXT_BLOCK_SIZE	0x100000
 static short
 ptp_read_func (unsigned char *bytes, unsigned int size, void *data, unsigned int *readbytes)
 {
@@ -294,19 +372,36 @@ ptp_read_func (unsigned char *bytes, unsigned int size, void *data, unsigned int
    */
   while (curread < size) {
     toread = size - curread;
-    if (toread > 4096)
-      toread = 4096;
+    if (toread > CONTEXT_BLOCK_SIZE)
+      toread = CONTEXT_BLOCK_SIZE;
     
     result = USB_BULK_READ(ptp_usb->handle, ptp_usb->inep,(char *)(bytes+curread), toread, ptpcam_usb_timeout);
     if (result == 0) {
       result = USB_BULK_READ(ptp_usb->handle, ptp_usb->inep,(char *)(bytes+curread), toread, ptpcam_usb_timeout);
     }
     if (result < 0)
-      break;
+      return PTP_ERROR_IO;
     curread += result;
     if (result < toread) /* short reads are common */
       break;
   }
+
+  // Increase counters, call callback
+  if (ptp_usb->callback_active) {
+    ptp_usb->current_transfer_complete += curread;
+    if (ptp_usb->current_transfer_complete > ptp_usb->current_transfer_total) {
+      // Fishy... but some commands have unpredictable lengths.
+      // send last update and disable callback.
+      ptp_usb->current_transfer_complete = ptp_usb->current_transfer_total;
+      ptp_usb->callback_active = 0;
+    }
+    if (ptp_usb->current_transfer_callback != NULL) {
+      (void) ptp_usb->current_transfer_callback(ptp_usb->current_transfer_complete,
+						ptp_usb->current_transfer_total,
+						ptp_usb->current_transfer_callback_data);
+    }
+  }
+  
   if (result > 0) {
     *readbytes = curread;
     return (PTP_RC_OK);
@@ -323,25 +418,64 @@ ptp_write_func (unsigned char *bytes, unsigned int size, void *data)
   int towrite = 0;
   int result = 0;
   int curwrite = 0;
-
+  
   /*
    * gp_port_write returns (in case of success) the number of bytes
-   * write. Too large blocks (>5x MB) could timeout.
+   * written. Too large blocks (>5x MB) could timeout.
    */
   while (curwrite < size) {
     towrite = size-curwrite;
-    if (towrite > 4096)
-      towrite = 4096;
+    if (towrite > CONTEXT_BLOCK_SIZE)
+      towrite = CONTEXT_BLOCK_SIZE;
     result=USB_BULK_WRITE(ptp_usb->handle,ptp_usb->outep,(char *)(bytes+curwrite),towrite,ptpcam_usb_timeout);
     if (result < 0)
-      break;
+      return PTP_ERROR_IO;
     curwrite += result;
     if (result < towrite) /* short writes happen */
       break;
   }
-  // Should load wMaxPacketsize from endpoint first. But works fine for all EPs.
-  if ((size % 512) == 0)
-    result=USB_BULK_WRITE(ptp_usb->handle,ptp_usb->outep,(char *)"x",0,ptpcam_usb_timeout);
+
+  // Increase counters, call callback
+  if (ptp_usb->callback_active) {
+    ptp_usb->current_transfer_complete += curwrite;
+    if (ptp_usb->current_transfer_complete > ptp_usb->current_transfer_total) {
+      // Fishy... but some commands have unpredictable lengths.
+      // send last update and disable callback.
+      ptp_usb->current_transfer_complete = ptp_usb->current_transfer_total;
+      ptp_usb->callback_active = 0;
+    }
+    if (ptp_usb->current_transfer_callback != NULL) {
+      (void) ptp_usb->current_transfer_callback(ptp_usb->current_transfer_complete,
+						ptp_usb->current_transfer_total,
+						ptp_usb->current_transfer_callback_data);
+    }
+  }
+  
+  // If this is the last transfer (callbacks only active if this function called repeatedly with
+  // new data, otherwise this is a single large transaction which ends here).
+  // a request and the bulk send header.
+  if (!ptp_usb->callback_active) {
+    // Then terminate an even packet boundary write with a zero length packet
+    if ((size % ptp_usb->outep_maxpacket) == 0) {
+      result=USB_BULK_WRITE(ptp_usb->handle,ptp_usb->outep,(char *)"x",0,ptpcam_usb_timeout);
+    }
+  } else if (ptp_usb->current_transfer_complete == ptp_usb->current_transfer_total) {
+    // This is the amount actually transfered in this large transaction.
+    uint64_t actual_xfer_size = ptp_usb->current_transfer_total - 2*PTP_USB_BULK_HDR_LEN;
+    
+    if ((actual_xfer_size % ptp_usb->outep_maxpacket) == 0) {
+      result=USB_BULK_WRITE(ptp_usb->handle,ptp_usb->outep,(char *)"x",0,ptpcam_usb_timeout);
+    }
+    // Set as complete and disable callback, just as good. 
+    // This also blocks callbacks from the following response command.
+    if (ptp_usb->current_transfer_callback != NULL) {
+      (void) ptp_usb->current_transfer_callback(ptp_usb->current_transfer_total, // Both total = 100%
+						ptp_usb->current_transfer_total, // Both total = 100%
+						ptp_usb->current_transfer_callback_data);
+    }
+    ptp_usb->callback_active = 0;
+  }
+
   if (result < 0)
     return PTP_ERROR_IO;
   return PTP_RC_OK;
@@ -388,6 +522,18 @@ static int init_ptp_usb (PTPParams* params, PTP_USB* ptp_usb, struct usb_device*
       return -1;
     }
     ptp_usb->handle = device_handle;
+#ifdef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
+    /*
+     * If this device is known to be wrongfully claimed by other kernel
+     * drivers (such as mass storage), then try to unload it to make it
+     * accessible from user space.
+     */
+    if (ptp_usb->device_flags & DEVICE_FLAG_UNLOAD_DRIVER) {
+      if (usb_detach_kernel_driver_np(device_handle, dev->config->interface->altsetting->bInterfaceNumber)) {
+	perror("usb_detach_kernel_driver_np()");
+      }
+    }
+#endif
     if (usb_claim_interface(device_handle, dev->config->interface->altsetting->bInterfaceNumber)) {
       perror("usb_claim_interface()");
       return -1;
@@ -489,6 +635,8 @@ uint16_t connect_first_device(PTPParams *params, PTP_USB *ptp_usb, uint8_t *inte
   uint16_t ret=0;
   int n;
 
+  // Reset device flags
+  ptp_usb->device_flags = DEVICE_FLAG_NONE;
   // First try to locate the device using the extended
   // device descriptor.
   dev = probe_usb_bus_for_mtp_devices();
@@ -503,6 +651,7 @@ uint16_t connect_first_device(PTPParams *params, PTP_USB *ptp_usb, uint8_t *inte
 	  dev->descriptor.idProduct == mtp_device->product_id ) {
 	printf("Autodetected device \"%s\" (VID=%04x,PID=%04x) is known.\n", 
 	       mtp_device->name, dev->descriptor.idVendor, dev->descriptor.idProduct);
+	ptp_usb->device_flags = mtp_device->device_flags;
 	break;
       }
     }
@@ -530,6 +679,7 @@ uint16_t connect_first_device(PTPParams *params, PTP_USB *ptp_usb, uint8_t *inte
 	      dev->descriptor.idProduct == mtp_device->product_id ) {
 	    
 	    printf("Found non-autodetected device \"%s\" on USB bus...\n", mtp_device->name);
+	    ptp_usb->device_flags = mtp_device->device_flags;
             goto next_step;
 	    
 	  }
@@ -544,11 +694,11 @@ next_step:
     return PTP_CD_RC_NO_DEVICES;
   }
 
-  // Found a device, then assign endpoints...
+  // Found a device, then assign endpoints to ptp_usb...
   ep = dev->config->interface->altsetting->endpoint;
   n = dev->config->interface->altsetting->bNumEndpoints;
-  find_endpoints(dev, &(ptp_usb->inep), &(ptp_usb->inep_maxpacket),
-		 &(ptp_usb->outep), &(ptp_usb->outep_maxpacket), &(ptp_usb->intep));
+  find_endpoints(dev, &ptp_usb->inep, &ptp_usb->inep_maxpacket,
+		 &ptp_usb->outep, &ptp_usb->outep_maxpacket, &ptp_usb->intep);
   
   // printf("Init PTP USB...\n");
   if (init_ptp_usb(params, ptp_usb, dev) < 0) {
@@ -598,7 +748,7 @@ static void find_endpoints(struct usb_device *dev, int* inep, int* inep_maxpacke
       if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)==0)
 	{
 	  *outep=ep[i].bEndpointAddress;
-	  *inep_maxpacket=ep[i].wMaxPacketSize;
+	  *outep_maxpacket=ep[i].wMaxPacketSize;
 	}
     } else if (ep[i].bmAttributes==USB_ENDPOINT_TYPE_INTERRUPT){
       if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)==
