@@ -23,6 +23,7 @@
 
 #include <stdarg.h>
 #include <time.h>
+#include <iconv.h>
 #include "gphoto2-endian.h"
 
 /* PTP datalayer byteorder */
@@ -182,8 +183,7 @@ typedef struct _PTPIPHeader PTPIPHeader;
 /* 9003 - sends nothing, nothing back  */
 /* no 9004 observed yet */
 /* no 9005 observed yet */
-/* 9006 - sends a ptp string, gets back a uint32 */
-#define PTP_OC_CANON_9006		0x9006
+#define PTP_OC_CANON_LookupObject	0x9006
 /* no 9007 observed yet */
 #define PTP_OC_CANON_StartShootingMode	0x9008
 #define PTP_OC_CANON_EndShootingMode	0x9009
@@ -722,7 +722,6 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_DTC_AINT128		(PTP_DTC_ARRAY_MASK | PTP_DTC_INT128)
 #define PTP_DTC_AUINT128	(PTP_DTC_ARRAY_MASK | PTP_DTC_UINT128)
 
-#define PTP_DTC_UNISTR		0xFFFE
 #define PTP_DTC_STR		0xFFFF
 
 /* Device Properties Codes */
@@ -814,7 +813,7 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_DPC_NIKON_RawCompression			0xD016
 #define PTP_DPC_NIKON_WhiteBalanceAutoBias		0xD017
 #define PTP_DPC_NIKON_WhiteBalanceTungstenBias		0xD018
-#define PTP_DPC_NIKON_WhiteBalanceFlourescentBias	0xD019
+#define PTP_DPC_NIKON_WhiteBalanceFluorescentBias	0xD019
 #define PTP_DPC_NIKON_WhiteBalanceDaylightBias		0xD01A
 #define PTP_DPC_NIKON_WhiteBalanceFlashBias		0xD01B
 #define PTP_DPC_NIKON_WhiteBalanceCloudyBias		0xD01C
@@ -1042,10 +1041,12 @@ typedef short (* PTPIOWriteFunc)(unsigned char *bytes, unsigned int size,
  */
 typedef uint16_t (* PTPIOSendReq)	(PTPParams* params, PTPContainer* req);
 typedef uint16_t (* PTPIOSendData)	(PTPParams* params, PTPContainer* ptp,
-					unsigned char *data, unsigned int size);
+					 unsigned char *data, unsigned int size,
+					 int from_fd);
 typedef uint16_t (* PTPIOGetResp)	(PTPParams* params, PTPContainer* resp);
 typedef uint16_t (* PTPIOGetData)	(PTPParams* params, PTPContainer* ptp,
-					unsigned char **data, unsigned int *recvlen);
+	                                unsigned char **data, unsigned int *recvlen,
+	                                int to_fd);
 /* debug functions */
 typedef void (* PTPErrorFunc) (void *data, const char *format, va_list args)
 #if (__GNUC__ >= 3)
@@ -1106,25 +1107,33 @@ struct _PTPParams {
 	uint8_t		cameraguid[16];
 	uint32_t	eventpipeid;
 	char		*cameraname;
+
+	/* iconv converters */
+	iconv_t cd_locale_to_ucs2;
+	iconv_t cd_ucs2_to_locale;
 };
 
 /* last, but not least - ptp functions */
 uint16_t ptp_usb_sendreq	(PTPParams* params, PTPContainer* req);
 uint16_t ptp_usb_senddata	(PTPParams* params, PTPContainer* ptp,
-				unsigned char *data, unsigned int size);
+				 unsigned char *data, unsigned int size,
+				 int from_fd);
 uint16_t ptp_usb_getresp	(PTPParams* params, PTPContainer* resp);
 uint16_t ptp_usb_getdata	(PTPParams* params, PTPContainer* ptp, 
-				unsigned char **data, unsigned int *readlen);
+				 unsigned char **data, unsigned int *readlen,
+	                         int to_fd);
 uint16_t ptp_usb_event_check	(PTPParams* params, PTPContainer* event);
 uint16_t ptp_usb_event_wait	(PTPParams* params, PTPContainer* event);
 
 int      ptp_ptpip_connect	(PTPParams* params, const char *port);
 uint16_t ptp_ptpip_sendreq	(PTPParams* params, PTPContainer* req);
 uint16_t ptp_ptpip_senddata	(PTPParams* params, PTPContainer* ptp,
-				unsigned char *data, unsigned int size);
+				unsigned char *data, unsigned int size,
+				int to_fd);
 uint16_t ptp_ptpip_getresp	(PTPParams* params, PTPContainer* resp);
 uint16_t ptp_ptpip_getdata	(PTPParams* params, PTPContainer* ptp, 
-				unsigned char **data, unsigned int *readlen);
+				 unsigned char **data, unsigned int *readlen,
+	                         int to_fd);
 uint16_t ptp_ptpip_event_wait	(PTPParams* params, PTPContainer* event);
 uint16_t ptp_ptpip_event_check	(PTPParams* params, PTPContainer* event);
 
@@ -1136,6 +1145,7 @@ uint16_t ptp_closesession	(PTPParams *params);
 uint16_t ptp_getstorageids	(PTPParams* params, PTPStorageIDs* storageids);
 uint16_t ptp_getstorageinfo 	(PTPParams* params, uint32_t storageid,
 				PTPStorageInfo* storageinfo);
+uint16_t ptp_formatstore        (PTPParams* params, uint32_t storageid);
 
 uint16_t ptp_getobjecthandles 	(PTPParams* params, uint32_t storage,
 				uint32_t objectformatcode,
@@ -1152,6 +1162,7 @@ uint16_t ptp_getobjectinfo	(PTPParams *params, uint32_t handle,
 
 uint16_t ptp_getobject		(PTPParams *params, uint32_t handle,
 				unsigned char** object);
+uint16_t ptp_getobject_tofd     (PTPParams* params, uint32_t handle, int fd);
 uint16_t ptp_getpartialobject	(PTPParams* params, uint32_t handle, uint32_t offset,
 				uint32_t maxbytes, unsigned char** object);
 uint16_t ptp_getthumb		(PTPParams *params, uint32_t handle,
@@ -1164,7 +1175,8 @@ uint16_t ptp_sendobjectinfo	(PTPParams* params, uint32_t* store,
 				uint32_t* parenthandle, uint32_t* handle,
 				PTPObjectInfo* objectinfo);
 uint16_t ptp_sendobject		(PTPParams* params, unsigned char* object,
-				uint32_t size);
+				 uint32_t size);
+uint16_t ptp_sendobject_fromfd  (PTPParams* params, int fd, uint32_t size);
 
 uint16_t ptp_initiatecapture	(PTPParams* params, uint32_t storageid,
 				uint32_t ofc);
@@ -1229,6 +1241,7 @@ uint16_t ptp_canon_getfolderentries (PTPParams* params, uint32_t store,
 				uint32_t handle, 
 				PTPCANONFolderEntry** entries,
 				uint32_t* entnum);
+uint16_t ptp_canon_lookup_object (PTPParams* params, char* name, uint32_t* objectid);
 uint16_t ptp_canon_theme_download (PTPParams* params, uint32_t themenr,
 				unsigned char **data, unsigned int *size);
 
